# Facer
# John W. Miller
# 2019

import sys
assert sys.version_info[0] == 3, "facer requires Python 3."

"""Simple Python face averaging"""

__author__ = 'John W. Miller'
__url__ = 'https://github.com/johnwmillr/facer'
__description__ = 'Face averaging in Python made simple'
__license__ = 'MIT'
__version__ = '0.6.0'

print("Loading face detector and landmark prediction models...")
